//---------------------------------------------------------------------------
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  1998-2001 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
//---------------------------------------------------------------------------
// stdmac.uc
// standard macros
//
//
// system: IXP1200

// Debugging in Hardware:
// 		Remove VALIDATION_CARD in build settings for microengines 2 and 3
//		This allows only 8 receive threads to run, for ports 0-7.

// Debugging in Simulation (present settings)
// 		Define VALIDATION_CARD in build settings for microengines 2 and 3.
//		This allows 16 receive threads to run, for ports 0-15.
// 
// modified for ECE697J Lab 2 Part II
	

//#define RFC1812
#include "project_config.h"
#include "mem_map.h"
#include "ixplib.uc"

// START ECE697J DEFINITIONS

#ifndef  ETHER_ARP
	#define  ETHER_ARP 0x806
#endif

#define TCP_PACKET 0x06 
#define TCP_SSH    0x16  //SSH: 22
#define TCP_WEB    0x50  //WEB: 80 


#define UDP_PACKET 0x11 //decimal 17 

#ifndef ETHER_PROTOCOL_LENGTH 
	#define  ETHER_PROTOCOL_LENGTH 				12, 2		// protocol length field byte offset 12, length 2 bytes
#endif

#ifndef IP_UPPER_LAYER_PROTOCOL
#define IP_UPPER_LAYER_PROTOCOL            7, 1         // Uppper layer protocl byte offeset 7(not 9 due to buffer starting the second byte of IP),  length 1 byte
																											
#endif	

#ifndef TCP_DEST_PORT 
	#define  TCP_DEST_PORT                      2, 2         // TCP destination port byte offset 2,   length 2 bytes	
#endif 

// END ECE697J DEFINITIONS

// initializations
sig_ifctx_block(0, inter_thread, OR, 1);		// block until signal global initialization complete

port_rx_init(rec_req, rfifo_addr, UENGINE_ID, PORT_NUMS, FAST_PORT_LO, FAST_PORT_HI);	// create a canned receive request
critsect_init(@rdready_inflight);			// init semaphore to limit 1 rec ready read per engine
critsect_init(@req_inflight);				// init semaphore to limit 1 rec req per engine

//immed32(route_base, SDRAM_ROUTE_FWD_BASE);					// init table addresses	
move(rec_state, 0);											// init rec_state
immed32(packet_buf_addr, UNALLOCATED);						// address to pop from freelist
xbuf_alloc($$dxfer, 8);					// allocate sdram xfers as a block of 8


// A validation card would have 16 100M ports. For simulation, define VALIDATION_CARD to enable
//	receive microengines 0-3, ports 0-15.
//
// the eval card only has 8 100M ports. For eval card hardware do not define VALIDATION_CARD.
//	This will enable receive icroengines 0-1, ports 0-7.
//
#ifndef VALIDATION_CARD		
#if (UENGINE_ID > 1)
	ctx_arb[kill]
	nop
	nop
	nop
#endif
#endif

// main loop   
.while (1)

	xbuf_alloc($pop_xfer, 1);

    .if (packet_buf_addr == UNALLOCATED)
    	buf_pop($pop_xfer[0], FREELIST_HANDLE, sig_done);							// if no buffer, get ptr
    .endif

	port_rxrdy_chk(@rdready_inflight, rec_req);
	critsect_enter[@req_inflight]						; block other contexts from sending rec_req
	port_rx_request(rec_req);							// get mpacket

    #ifdef RFC1812
   		port_rx_receive(exception, recv_port, rec_state, ETHER_100M);	// get mpacket status
    #else
	    port_rx_receive(exception, rec_state, ETHER_100M);	// get mpacket status
    #endif //RFC1812

	mpacket_received#:
																				// wait for mem_pop to complete
    .if (packet_buf_addr == UNALLOCATED)
		buf_wait();
		#if (FREELIST_ID == 0)
			#define BASE_ADDR  SRAM_BUFF_DESCRIPTOR_BASE
		#else 
			#define_eval HALF_BUFFER_COUNT (BUFFER_COUNT / 2)
			#define_eval BASE_ADDR (SRAM_BUFF_DESCRIPTOR_BASE + (HALF_BUFFER_COUNT * 4))
		#endif
	    .while ($pop_xfer[0] == BASE_ADDR)
			buf_pop($pop_xfer[0], FREELIST_HANDLE, ctx_swap);					// if no buffer available, retry
		.endw
		buf_dram_addr_from_sram_addr(packet_buf_addr, $pop_xfer[0], FREELIST_HANDLE);		// calc packet buffer address
		move(descriptor_addr, $pop_xfer[0]);
	.endif
	xbuf_free($pop_xfer);

	.if (exception != 0)
		.if (exception == PORT_RXFAIL)
			port_rx_fail(rec_state, rec_req, exception);							// fail is like EOP w/o queueing
			buf_dram_addr_from_sram_addr(packet_buf_addr, descriptor_addr, FREELIST_HANDLE);		// calc packet buffer address
			critsect_exit[@req_inflight]						; allow other contexts to send rec_req
			.continue
		.elif (exception == PORT_RXERROR)
			port_rx_error(rec_state, rec_req, exception);							// bump mib counter
			buf_dram_addr_from_sram_addr(packet_buf_addr, descriptor_addr, FREELIST_HANDLE);		// calc packet buffer address
			critsect_exit[@req_inflight]						; allow other contexts to send rec_req
			.continue
		.endif
	.endif

	critsect_exit[@req_inflight]						; allow other contexts to send rec_req

	.if (!(1 & (rec_state >> REC_STATE_SOP_BIT)))              // if not SOP
		port_rx_bytecount_extract(current_bytecount, rec_state);				// update packet byte count
		shf_right(current_qwcount, current_bytecount, 3);
		add(packet_buf_addr, packet_buf_addr, QWCOUNT8);
		dram_rfifo_read(packet_buf_addr, QWOFFSET0, rfifo_addr, QWOFFSET0, current_qwcount, ctx_swap);// write mpacket to dram from RFIFO
	.elif (bit(rec_state, REC_STATE_SOP_BIT))		// if SOP
	     xbuf_alloc($pkt_buf, 4);													// allocate xfer buffer 4 lwords
		 nop

     #ifdef RFC1812 // Initialize Port Enable/Disable, Promiscous/Nonpromiscous from scratch
    .local temp 
	     immed32[temp, SCRATCH_IF_ENABLED]
         scratch[read, $pkt_buf0,temp, 0x0, 2], ctx_swap
	     alu[PMode, --, b, $pkt_buf1]
	     alu[port_status, --, b, $pkt_buf0]
    .endlocal
    #endif //RFC1812


		 rfifo_read($pkt_buf[0], rfifo_addr, QWOFFSET0, QWCOUNT2, ctx_swap);				// read link layer, 1st 2 quadwords

         #ifdef RFC1812 //Check whether the receive port is enabled [RFC1812,5.3.11]
         port_enabled_chk(exception, recv_port, port_status) 
		.if (exception == PORT_DISABLED)
		   port_rx_discard(rec_state, rec_req, exception);				// set discard, inc counter
		   br[freebuf#]
        .endif
        #endif //RFC1812

		xbuf_extract(protocol_len, $pkt_buf, BYTEOFFSET0, ETHER_PROTOCOL_LENGTH); // get protocol length field
	
	   #ifdef RFC1812

       #else
#ifdef LITTLE_ENDIAN
		  shf_right(sa01, $pkt_buf[1], 16);
#else
		  alu_op(sa01, 0, +16, $pkt_buf[1]);										// save sa bytes 0-1 for later merge
#endif //LITTLE ENDIAN

      #endif //RFC1812
		 	
		 .if (protocol_len > ETHER_MTU)				// if NOT 802.3  (see RFC 1122, 2.3.3)

             #ifdef RFC1812
		     .if (protocol_len != ETHER_ARP)  ;Arp Packets send to IP Stack (rfc1812 3.3.2)
     		     ether_addr_validation(exception, $pkt_buf0,$pkt_buf1);	; verify l2 destn address (rfc1812 5.3.4)
	             .if ((exception != ETHER_BROADCAST) && (exception != ETHER_MULTICAST))
				    port_promiscous_chk(exception, recv_port, PMode) ;verify whether mac filtering is enabled
					.if (exception) //Filter mode
				        ether_addr_filter(exception, $pkt_buf[0], $pkt_buf[1],recv_port) ;MAC address filtering
				       .if (exception == MAC_INVALID_ADDRESS)
   					       port_rx_discard(rec_state, rec_req, exception);				// set discard, inc counter
					       br[freebuf#]
				       .endif
                    .endif
             #endif //RFC1812

				// START ECE697J Identification of non-IP packets
			 	.if (protocol_len != ETHER_ARP) 
				// END ECE697J Identification of non-IP packets
					
			     	xbuf_extract(ip_verslen, $pkt_buf, BYTEOFFSET14, IP_VERSION_LENGTH); // get version/length field
		
					.if (ip_verslen == IPV4_NO_OPTIONS)		// if ipv4 no options		// see RFC 791

				        xbuf_alloc($pkt_buf_ip, 4);										// allocate xfer buffer 4 lwords
				        xbuf_link($pkt_buf, $pkt_buf_ip);								// linke the xfer buffers
				        xbuf_link($pkt_buf_ip, $pkt_buf);								// create circular buffer

				        dram_rfifo_read(packet_buf_addr, QWOFFSET4, rfifo_addr, QWOFFSET4, QWCOUNT4, sig_done);// write 2nd 32B
    			        rfifo_read($pkt_buf_ip[0], rfifo_addr, QWOFFSET2, QWCOUNT3, ctx_swap);	// read next 3 quadwords

						// START ECE 697J CLASSIFICATION
						xbuf_extract(ip_upp_pro, $pkt_buf_ip, BYTEOFFSET0, IP_UPPER_LAYER_PROTOCOL); // read protocol number from IP header
						.if (ip_upp_pro == TCP_PACKET)
					  		xbuf_extract(tcp_dport, $pkt_buf_ip, BYTEOFFSET18, TCP_DEST_PORT); 	 	// read TCP destination port number
							.if (tcp_dport == TCP_SSH)
							   move(output_intf, 0x0000008);
							.else 
							   move(output_intf, 0x00000008);		
							.endif	
						.else 
						 	move(output_intf, 0x00000000);
		                .endif 
					 	//move(output_intf, 0x00000000); // default case - just in case
						// END ECE 697J CLASSIFICATION

                        #ifdef RFC1812
					    	ip_filter(exception, $pkt_buf);
							.if ((exception != IGMP_PROTOCOL) && (exception != SNMP_PROTOCOL))	 
						 	.local ip_sa
						    	xbuf_extract(ip_sa, $pkt_buf, BYTEOFFSET14, IP_SOURCE_ADDRESS); 	 	// get src ip address
	                        	xbuf_extract(ip_da,$pkt_buf, BYTEOFFSET14, IP_DESTINATION_ADDRESS); 	// get src ip address
	                            ctx_arb[voluntary]
							    ip_addr_validation(exception, ip_sa);			// check ip address
								.if (exception !=0x0)
								    br[packet_late_discard#] ; drop/discard packet if ip dst is dbcast.
								.endif
								// Ensure both ip src and dst addresses
	            	            //  aren't a directed broadcast address.
								ip_dbcast_check(exception, ip_sa,$pkt_buf2);  
								.if (exception == IP_DBCAST_ADDRESS)
	                        	    br[packet_late_discard#] ; drop/discard packet if ip dst is dbcast.
	                            .endif
		                        alu[exception, --, B, 0]   ; OK - ip src and dst aren't d
							.endlocal
            	            #endif //RFC1812

				 			ip_verify(exception, $pkt_buf, BYTEOFFSET14);					// check ip header
				            ctx_arb[voluntary]
				            ip_modify($$dxfer, BYTEOFFSET14, $pkt_buf, BYTEOFFSET14);			// modifiy ip packet

							#ifdef RFC1812
							    xbuf_extract(pkt_len, $pkt_buf, BYTEOFFSET14, IP_TOTAL_LENGTH); // get total length field
                            #else //RFC1812
					            xbuf_extract(ip_da, $pkt_buf, BYTEOFFSET14, IP_DESTINATION_ADDRESS); // get ip dest field
                            #endif

				            xbuf_free($pkt_buf_ip);

                            #ifdef RFC1812 //(RFC1812, 4.2.2.9 and 5.2.7.3)
					        .if ((exception) && (exception != IP_BAD_TTL))
              	            #else
   	          	           	.if (exception)
                            #endif //RFC1812

								packet_late_discard#:
								port_rx_discard(rec_state, rec_req, exception);				// set discard, inc counter
					            sig_wait(sdram, OR, 1);
				        	.else	//???????
                                							
							  #ifdef RFC1812
							  	.if ($$dxfer[0] == DBCAST_PORT) ;check directed broadcast for the destination address
							    	port_rx_discard(rec_state, rec_req, exception);				// set discard, inc counter
								  	br[freebuf#]
							  	.endif
							  	.if ($$dxfer[0] == IP_INVALID_PORT) ;invalid forwarding port
							    	port_rx_discard(rec_state, rec_req, exception);				// set discard, inc counter
								  	br[freebuf#]
							  	.endif
                              	.if (exception == IP_BAD_TTL) // if ttl exception and the destination is the host send it to IP Stack (4.2.2.9)
					            	.if ($$dxfer[0] == CORE_STACK_INTF1) 
		                            	dram_rfifo_read(packet_buf_addr, QWOFFSET0, rfifo_addr, QWOFFSET0, QWCOUNT8, ctx_swap);
                                   		br[freebuf#]
                                  	.endif
                                  	port_rx_discard(rec_state, rec_req, exception);				// set discard, inc counter
								 	br[freebuf#]
                              	.endif  //->if ttl exception and the destination is the host
							   	ctx_arb[voluntary]
							   	port_enabled_chk(exception, $$dxfer[0], port_status);
							   	.if (exception == PORT_DISABLED) //output port disabled discard packet
	                            	port_rx_discard(rec_state, rec_req, exception);				// set discard, inc counter
									br[freebuf#]
								.endif
							  	.if ($$dxfer0 != recv_port)
                              	.local mtu  //[RFC1812 3.3.4]
                                	// for fragmentation: Read mtu corresponding to the output port from memory
                                  	// Compare with the lenth of the packet, if the packet length is greater send 
                                  	// to strong arm for fragmentation.
			                      	ld_field_w_clr[mtu, 0011,$$dxfer7]
							      	.if (pkt_len <= mtu)
                              	#endif //RFC1812

					 			move($$dxfer[0], $$dxfer[1]);						// new DA bytes 0-3

								#ifdef RFC1812                        
									// Modify the outgoing packets source MAC address with the  
                                	// MAC address of the transmit port 
                                  	.local tmp1 tmp2
                                    	alu[tmp1, --,b,$$dxfer6,>>16]
                                      	alu[$$dxfer1, tmp1, +,$$dxfer2,<<16]
                                      	alu[tmp2, --,b, $$dxfer7,>>16]
                                      	alu[tmp1, --,b,$$dxfer6]
                                      	alu[$$dxfer2, tmp2, +, tmp1, <<16]                     ; 
                                   	.endlocal      

								#else	; NOT RFC1812
									#ifdef LITTLE_ENDIAN	
					                	add_shf_left($$dxfer[1], $$dxfer[2], sa01, 16);				// merge new DA 4-5 with SA 0-1
							      	#else
					                   add_shf_left($$dxfer[1], sa01, $$dxfer[2], 16);				// merge new DA 4-5 with SA 0-1
								  	#endif
			                        move($$dxfer[2], $pkt_buf[2]);									// previous SA bytes 2-5
                              	#endif //RFC1812


								dram_write($$dxfer[0], packet_buf_addr, QWOFFSET0, QWCOUNT4, no_signal);	// write modified packet

								#ifdef RFC1812
                              			br[freebuf#]		; execution optimization. avoid a bunch of branches
									.else // Packet length is greater than mtu send to core for fragmentation
				      					packetq_core(packet_buf_addr, output_intf) // if not ethernet, q to core								
									.endif //  -> 	.if (pkt_len <= mtu)
									.endlocal  // mtu ->	.local mtu  //[RFC1812 3.3.4]
                            	.else // Receive port and transmit port are the same, send to the core for SNMP redirect message
			       					packetq_core(packet_buf_addr, output_intf) // if not ethernet, q to core								
							   	.endif // -> 	.if ($$dxfer0 != recv_port)
                               	#endif //RFC1812

				         .endif // LOOK-UP

                     #ifdef RFC1812
				     .else					//ICMP, IGMP or SNMP packets--
	     				 packetq_core(packet_buf_addr, output_intf) // if not ethernet, q to core-->IGMP 								
			         .endif	 								
                     #endif //RFC1812

			    .else						// IP with options or frag
				     dram_rfifo_read(packet_buf_addr, QWOFFSET0, rfifo_addr, QWOFFSET0, QWCOUNT8, ctx_swap);
				     shf_left(output_intf, CORE_STACK_INTF1, 3);							// if options or frag, q to core
			    .endif								// IP with options or frag

			
             #ifdef RFC1812
             .else
				 packetq_core(packet_buf_addr, output_intf) // if not ethernet, q to core-->not Unicast								
	         .endif // L2 Broadcast or Multicast
          .else
  										
          .endif //Arp packets --> protocol_len !=ETHER_ARP
          #endif //RFC1812
			
			// START ECE697J Handling of non-IP packets
			.else
				 
				move(output_intf, 0x00000000);
			.endif 
			// END ECE697J Handling of non-IP packets


		.else										// not ethernet -->if (protocol_len > ETHER_MTU)
			dram_rfifo_read(packet_buf_addr, QWOFFSET0, rfifo_addr, QWOFFSET0, QWCOUNT8, ctx_swap);
			shf_left(output_intf, CORE_STACK_INTF1, 3);								// if not ethernet, q to core
		.endif									// not ethernet --->if (protocol_len > ETHER_MTU)
 	.endif										// not SOP and SOP




freebuf#:

	xbuf_free($pkt_buf);


	.if (bit(rec_state, REC_STATE_EOP_BIT))			// if EOP
		.if (!bit(rec_state, REC_STATE_DISCARD_BIT))		// if not discard
			xbuf_alloc($desc_buf, 2);
			move($desc_buf[0], 0);														// descriptor to xfer
			.local com_descriptor
				shf_left(com_descriptor, FREELIST_ID, DESC1_FREELIST);					// insert freelist id
                #ifdef RFC1812
					//Set receive port to descriptor entry, required for PETH driver to work
					alu[com_descriptor, com_descriptor, OR, recv_port, <<DESC1_RCV_PORT]
                #endif //RFC1812
				#if (DESC1_PKT_END_BYTE <= REC_STATE_BYTE_COUNT);
					#define_eval _SHIFT_AMT  (REC_STATE_BYTE_COUNT - DESC1_PKT_END_BYTE)
					field_shf_right_insert(com_descriptor, 0010, rec_state, _SHIFT_AMT);// insert last byte count
				#else
					#define_eval _SHIFT_AMT  (DESC1_PKT_END_BYTE - REC_STATE_BYTE_COUNT)
					field_shf_left_insert(com_descriptor, 0010, rec_state, _SHIFT_AMT);
				#endif

				alu_op(packet_buf_addr, QWCOUNT8, +8, packet_buf_addr);					// incr ele countex
				alu_shf_right($desc_buf[1], com_descriptor, +8, packet_buf_addr, 3); 	// insert element count
			.endlocal
			sram_write($desc_buf[0], descriptor_addr, LWOFFSET0, LWCOUNT2, ctx_swap);	// write descriptor

			enqueue#:

			#ifdef RFC1812 //Validate packet Size
				ip_pktvalidate_size[output_intf,packet_buf_addr,rec_state, pkt_len, exception] ;Validate packet size
	    		.if (exception == IP_BAD_TOTAL_LENGTH)
		    		port_rx_discard(rec_state, rec_req, exception);
		    		br[freebuf#]			 ; If packet size does not match discard the packet
          		.endif
			#endif//RFC1812

			packetq_send(descriptor_addr, output_intf, Q_HANDLE); 						// at eop, q to xmit
			immed32(packet_buf_addr, UNALLOCATED);
			xbuf_free($desc_buf);
		.else											// Discard
			// set packet_buf_addr to begining of the packet - Reuse the packet
			 buf_dram_addr_from_sram_addr(packet_buf_addr, descriptor_addr, FREELIST_HANDLE);
		.endif												// not discard
		move(rec_state, 0);															// init rec_state
	.endif											// EOP



header_stored#:
mid_mpacket_stored#:
last_mpacket_stored#:
.endw
xbuf_free($$dxfer);
nop



